/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created
    Nov 30 2000 Richard Jones   Generate separate files rather than inner classes

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */

package MetaTF.Generator.Java;

import java.io.*;
import MetaTF.Parser.DTD;
import MetaTF.Panic; 

/**
 * CodeGenerator.
 * Generate Java reader and writer
 *
 * @see <a href="../../Parser/metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.2 $, $Date: 2000/12/15 16:29:11 $),
 * @author Richard Jones
**/
public class CodeGenerator {
    private static final String cvsheader = "$Header: /home/cur/rej/cvs/MetaTF/src/MetaTF/Generator/Java/CodeGenerator.java,v 1.2 2000/12/15 16:29:11 rej Exp $";
    
    /** The name of the dtd **/
    private String dtd;
    
    /** The name of the class to be generated **/
    private String className;

    /** The parent directory **/
    private File parentDir;

    /** The FileWriter to which the class is to be written **/
    private FileWriter writer;

    /**
     * Allocate a new object representing the code generator
     * (should this extend / implement something?)
     *
     * @param dtdFile the name of the DTD
     * @param classname the name of the class file to be generated
    **/
    public CodeGenerator(String dtdFile, String className) {
        dtd = dtdFile;

        File tmp = new File(className);
        this.className = Util.classify(tmp.getName());
        parentDir = new File(tmp.getParent(), this.className);
        if(parentDir.exists()) 
            Panic.warning("Overwriting "+parentDir.getAbsolutePath());
        else if(!parentDir.mkdir()) 
            Panic.error("Cannot create directory "+parentDir.getAbsolutePath());
        try {
            writer = new FileWriter(new File(parentDir,className+".java"));
        } catch(IOException e) {
            Panic.error(e.getMessage());
        }
    }


    /**
     * Generate the code.
     *
     * @param symtab the symbol table
     * @param data not used
    **/
    public Object generate(DTD symtab, Object data) {

        /*
         * 1. Generate main getNextRecord loop and const names for fields
         */
        out(
"package "+className+";                                    \n\n"+        
"import java.util.*;                                       \n"+
"import java.io.*;                                         \n"+
"import MetaTF.BadRecordException;                         \n"+
"import MetaTF.MetaTFConstants;                            \n"+
"import MetaTF.Runtime.*;                                  \n\n"+
"/**                                                       \n"+
" * Reader/writer class for MetaTF DTD "+dtd+"             \n"+
" *  generated by MetaTFtool 1.2                           \n"+
"**/                                                       \n"+
"public class "+this.className+" {                         \n");
        MainGenerator mg = new MainGenerator(writer);
        symtab.accept(mg, null);

        /*
         * 2. Generate code to initialise Trace data structures
         */
        DataStructureGenerator dsg = new DataStructureGenerator(className, writer);
        symtab.accept(dsg, null);
        out(
"}                                                         ");

        /* Finished generating <class>.java */
        try {
            writer.flush();
            writer.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
        }


        /*
         * 3. Generate classes for each user-defined section and record
         */
        ClassGenerator cg = new ClassGenerator(dtd, parentDir, className);
        symtab.accept(cg, null);

        return null;
    }

    /**
     * Utility to emit code
     *
     * @param ln one or more lines of code
    **/
    private void out(String ln) {
        try {
            writer.write(ln+"\n"); 
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
    }

}// end of class CodeGenerator

