/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created
    Nov 30 2000 Richard Jones   Generate separate files rather than inner classes
 */

package MetaTF.Generator.Java;

import MetaTF.Parser.*;
import MetaTF.Types;
import MetaTF.Panic;
import MetaTF.Generator.Java.Util;
import java.util.*;
import java.io.*;

/**********************************************************************************************
 * MainGenerator.
 * Generate the code for the main interface
 *     constants encoding all field names
 *     public Record getNextRecord() throws BadRecordException, IOException;
 *
 * @see <a href="../../Parser/metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.4 $, $Date: 2001/03/29 15:53:13 $),
 * @author Richard Jones
**/
class MainGenerator implements DTDVisitor {

    /** The name of the section being visited **/
    String sectionName=null;

     /** The number of the section being visited **/
    int sectionNumber = 0;
    
    /** The TraceField number **/
    int fno=0;

    /** full name of field, e.g. "r.comment.length" **/
    String fullname;

    /** Suffix is needed by repeat fields to hold index **/
    String suffix;

    /** The symbol table **/
    DTD symtab;


    /** The FileWriter to which output is to be written **/
    private final FileWriter writer;

    /**
     * Allocate a new object to write putRecord methods
     *
     * @param w the FileWrite to which the method definition is to be written
    **/
    MainGenerator(FileWriter w) {
        writer = w;
        if(w==null)
            Panic.panic("out: writer must be called before out is called");
    }
 
    /**
     * Utility to emit code
     *
     * @param ln one or more lines of code
    **/
    private void out(String ln) {
        try {
            writer.write(ln+"\n"); 
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
    }


    /**
     * @param symtab the root of the symbol table
     * @param data not used
     * @return null
    **/
    public Object visit(DTD symtab, Object data) {

        this.symtab = symtab;

        /*
         * Emit constants for field names to simplify the API
         */
        out(        
"    // CONSTANTS FOR FIELD NAMES                                                         \n");
        Iterator it = symtab.getFields().iterator();
        for(int i=0; it.hasNext(); i++) {
            String fname = (String)it.next();
            out(
"    /** Encoding for all fields named "+fname+" **/                                      \n"+
"    public static final int "+Util.constantify(fname)+" = "+i+";                         ");
        }


       /*
        * Emit getNextRecord method
        * 1. Read the tag
        * 2. Try the specials
        *    For metadata records, we either
        *    - changeInterpretation of all given fields in all records in current section, or
        *    - set the encoding of an individual method
        * 3. Try records visible in the current section
        */
       out(
"\n    /**                                                                           \n"+
"     * Get the next record from the trace file.                                     \n"+
"     *                                                                              \n"+
"     * @exception IOException IF the read fails                                     \n"+
"     * @exception BadRecordException IF a badly-formed trace record is discovered   \n"+
"     *                               or any encoding problem is encountered   \n"+
"    **/                                                                             \n"+
"    public static Record getNextRecord(LowLevelReader llr) throws BadRecordException, IOException {   \n"+
"        TraceRecord tr;                                                             \n"+
"        TraceField tf;                                                              \n\n"+

"        // Section or Metadata?                                                     \n"+
"        int tag = MetaTFState.readSpecial(llr);                                     \n\n"+

"        // Comment?                                                                 \n"+
"        if(tag==MetaTFConstants.CommentTag) {                                       \n"+
"            Comment r = new Comment();                                              \n"+
"            r.text = MetaTFState.readln(llr);                                       \n"+     
"            return r;                                                               \n"+
"        }                                                                           \n\n"+

"          /*                                                                        \n"+
"           * Try records visible in current section                                 \n"+
"           * (defined either here or in section 0)                                  \n"+
"           */                                                                       \n"+
"          else switch(MetaTFState.getCurrentSection(tag)) {                         \n");

        Map sections = symtab.getSections();
        it = new TreeSet(sections.keySet()).iterator();
        while(it.hasNext()) {
            Integer sno = (Integer)it.next();
            ((Section)sections.get(sno)).accept(this, data);
            out("");
        }        
        out(
"            default: throw new BadRecordException(\"No section \"+MetaTFState.getCurrentSection(tag));   \n"+
"        }                                                                          \n"+
"    }                                                                              \n\n"+

"    /**                                                                            \n"+
"     * Write section metadata to change current section                            \n"+
"     * after changing the current section in the reader/writer                     \n"+
"     *                                                                             \n"+
"     * @param i the new section which must be valid                                \n"+
"     * @exception BadRecordException IF i is an invalid section number             \n"+
"     * @exception IOException IF the write fails                                   \n"+
"    **/                                                                            \n"+
"    public static void putSection(LowLevelWriter llw, int i) throws BadRecordException, IOException {  \n"+  
"        MetaTFState.putSection(llw, i);                             \n"+
"    }                                                                              \n\n"+

"    /**                                                                            \n"+
"     * Change the current section in the reader/writer                             \n"+
"     *                                                                             \n"+
"     * @param i the new section which must be valid                                \n"+
"     * @exception BadRecordException IF i is an invalid section number             \n"+
"    **/                                                                            \n"+
"    static void setSection(int i) throws BadRecordException {               \n"+   
"        MetaTFState.setSection(i);                                                 \n"+                     
"    }                                                                              \n");

        return null;
    }
        
    public Object visit(Section section, Object data) {
        sectionNumber = section.getNumber();
        sectionName = section.getName();
        out(
"            case "+sectionNumber+": // SECTION "+Util.classify(sectionName));

        /*
         * Try section Global 0, which is not cached
         */
        if(sectionNumber==0) {
            out(
"                if(MetaTFState.getTraceRecord(0,tag)==null)                        \n"+
"                    throw new BadRecordException(\"No record with tag\"+tag+\" (in section "+sectionNumber+")\"); \n"+
"                else switch(tag) {                                                 ");

            Map records = section.getRecords();
            Set keys = new TreeSet(records.keySet());
            Iterator it = keys.iterator();
            while(it.hasNext()) {
                String s = (String)it.next();
                Record r = (Record)records.get(s);
                if(r!=null && r.isTagged()) {
                    r.accept(this, Boolean.FALSE);
                }
                out("");
            }
        } else {
            out(
"                switch(tag) {                                                      ");

            Map records = section.getRecords();
            Set keys = new TreeSet(records.keySet());
            Iterator it = keys.iterator();
            while(it.hasNext()) {
                String s = (String)it.next();
                Record r = (Record)records.get(s);
                if(r!=null && r.isTagged()) {
                     r.accept(this, new Boolean(section.isCached()));
                }
                out("");
            }
        }
        out(
"                    default:                                                       \n"+
"                        throw new BadRecordException(\"No record with tag\"+tag+\" (in section "+sectionNumber+")\"); \n"+
"                }// end of records switch for section "+sectionNumber+" "+Util.classify(sectionName)+ "\n");

        return null;
    }


    /*
     * Visit this record
     *
     * visit Section ensures that this record has a tag
     * @param isCached should the record read be added to the section's cache?
     */
    public Object visit(Record record, Object isCached) {
        int recordTag;
        try {
            recordTag = record.getTag();
        } catch (NoTagException e ) {
            Panic.panic("MainGenerator.visit gets record with no tag, "+e.getMessage());
            return null;
        }
        out(
"                    case "+recordTag+": { // RECORD "+record.getName()                                     );

        String recordClass = Util.classify(record.getName());
        fullname = "r";
        suffix = "";
        String sectionRecordName = Util.classify(sectionName)+"_"+recordClass;
        out(
"                        "+sectionRecordName+" r = new "+sectionRecordName+"();                             \n"+
"                        tr = MetaTFState.getTraceRecord("+sectionNumber+", "+recordTag+");                      \n"+
"                        // process each field                                                              \n"+
"                        //r.tag = "+recordTag+";  // tags are final                                        ");

        Map knownFields = new HashMap();
        Field f = record.getField();
        if(f instanceof TagField) {
            return null;
        } else if(f instanceof NamedField) {
            Panic.panic("Found a 'tagged' record that holds only a NamedField, "+record.getName());
            return null;    // NEVER REACHED
        } else if(f instanceof RepeatField) {
            Panic.panic("Found a 'tagged' record that holds only a RepeatField, "+record.getName());
            return null;    // NEVER REACHED
        } else { // f instanceof TupleField - must treat outer-level tuples specially
            Iterator it = ((TupleField)f).getFields().iterator();
            if(it.hasNext())    // it must have the tag
                it.next();      // so consume it
            fno = 1;
            while(it.hasNext()) {
                ((Field)it.next()).accept(this, knownFields);
            }
        }

        if(((Boolean)isCached).booleanValue())
            out(
"                        "+Util.classify(sectionName)+".cache(r);                                         ");            
        out(
"                        return r;                                                                          \n"+
"                    } // end of case RECORD "+record.getName()+"                                           ");

        return null;
    }
        
    public Object visit(NamedField field, Object knownFields) {
        String name = Util.declassify(field.getName(), (Map)knownFields);
        String type = field.getType();
        String kind = Util.kindify(type);
        out(
"                        tf = tr.getTraceField("+fno+");                                                        ");
        if(Types.primitiveType(type)) {
            if(type.equals("long")) {
                out(
"                        "+fullname+"."+name+suffix+" = MetaTFState.readNum(llr, (NumTraceField)tf);            ");
            } else {
                out(
"                        "+fullname+"."+name+suffix+" = ("+type+")MetaTFState.readNum(llr, (NumTraceField)tf);  ");
            }
        } else if(type.equals("index")) {
            out(
"                        "+fullname+"."+name+suffix+" = MetaTFState.readIndex(llr, (IndexTraceField)tf);         ");
        } else if(type.equals("string")) {
            out(
"                        "+fullname+"."+name+suffix+" = MetaTFState.readString(llr, (StringTraceField)tf);       ");
        } else {
            Panic.panic("NamedField "+name+" gets unexpected type "+type+".");
	}

        fno++;
        return name;
    }
    
    public Object visit(TagField field, Object knownFields) {
        out(
"                        tf = tr.getTraceField("+fno+");                                                    \n"+
"                        /*"+fullname+".tag"+suffix+" = (int)MetaTFState.readNum(llr, (NumTraceField)tf);   tags are final*/ ");

        fno++;
        return "tag";
    }
    
    public Object visit(RepeatField field, Object knownFields) {
        String repeat = (String)field.getRepeat().accept(this, knownFields);
        String size = fullname+"."+repeat;
        Field repeatable = field.getRepeatable();
        String repeatableName = repeatable.getName();
        out(
"                        "+fullname+"."+repeatableName+" = new "+getType(repeatable)+"["+size+"];          \n"+
"                        for(int i=0; i<"+size+"; i++) {                                                   ");

        String oldSuffix = suffix;
        suffix = "[i]";
        field.getRepeatable().accept(this, knownFields);
        suffix = oldSuffix;
        out(
"                        }                                                                                 ");

        return null;
    }

    /*
     * Get the type of a field
     *
     * @param field the field
     */
    private String getType(Field field) {
        if(field instanceof TagField) 
            return "int"; 
        if(field instanceof NamedField) 
            return Util.typify(((NamedField)field).getType());
        if (field instanceof TupleField) {
            TupleField f = (TupleField) field;
            try {
                String name = f.getName();
                String cname
                     = symtab.getSection(sectionNumber).getRecords().containsKey(name) ?
                           Util.classify(symtab.getSection(sectionNumber).getName())+"."+Util.classify(name) :
                           Util.classify(symtab.getSection(0).getName())+"."+Util.classify(name);
                return cname;
            } catch (DTDException e) {
                // cannot happen
                Panic.panic(e.getMessage());
                return null;
            }
        } else {
            Panic.panic("getType finds a RepeatField");
            return null;
        }
    }
    
    public Object visit(TupleField field, Object knownFields) {
        String oldFullname = fullname;                              // e.g. "r"
        String name = field.getName();
        fullname = fullname+"."+Util.declassify(name, (Map)knownFields)+suffix; // e.g. "r.comment[i]"
        String oldSuffix = suffix;
        suffix="";        
        out(
"                        "+fullname+" = new "+getType(field)+"();                                                   ");

        Iterator it = field.getFields().iterator();
        while(it.hasNext()) {
            Field f = (Field)it.next();
            f.accept(this, new HashMap());  // Fields of the sub-record are local to it
        }
        suffix = oldSuffix;
        fullname = oldFullname;
        return null;
    }

    // NOT USED
    public Object visit(BaseOffset baseOffset, Object data)  { return null; }
    public Object visit(Default def, Object data)  { return null; }
    public Object visit(StringDefault def, Object data)  { return null; }
    public Object visit(Delta delta, Object data)  { return null; }
    public Object visit(None none, Object data)  { return null; }
    public Object visit(NumberValue numberValue, Object data)  { return null; }
    public Object visit(SectionStride sectionStride, Object data)  { return null; }
    public Object visit(Stride stride, Object data)  { return null; }
    public Object visit(SectionOffset value, Object data)  { return null; }
  

}// end of class MainGenerator

