
/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */
 
package MetaTF.Parser;

import java.util.*;
import MetaTF.Panic;

/**
 * Record
 *
 * @see <a href="metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.3 $, $Date: 2001/03/29 15:53:18 $),
 * @author Richard Jones
**/
public class Record implements HasFields, HasProperties, Comparable {
  private static final String cvsheader = "$Header: /home/cur/rej/cvs/MetaTF/src/MetaTF/Parser/Record.java,v 1.3 2001/03/29 15:53:18 rej Exp $";

    /** The name of the record **/
    private final String name;

    /** The record's (single) field **/
    private Field field;

    /** The section in which this record is defined **/
    private Section section;

    /** The line in the DTD where this record is defined **/
    private int beginLine;


    /**
     * Allocate a new object representing a record
     *
     * @param s the name of the record
     * @param l the line in the DTD where this record's definition started
     */
    public Record(String s, int l) {
        name = s;
        beginLine = l;
    }

    /**
     * Allocate a new object representing a record
     *
     * @param s the name of the record
     */
    Record(String s) {
        this(s, 0);
    }

    /**
     * Do a macro replacement of any fields that do not have
     * properties defined locally.
     *
     * @param inheritedRecords0 Map <name, Record> of records inherited from section 0
     * @param inheritedRecords Map <name, Record> of records inherited from this section
     * @param inheritedFields Map <name, Field> of inherited fields
    **/
    void replaceFields(Map inheritedFields0, Map inheritedRecords, Map inheritedFields) {
        if(field!=null)
    	    field = field.replaceFields(inheritedFields0, inheritedRecords, inheritedFields);
    	else
    	    Panic.panic(name+".replaceFields: field is null!!");
    }

    /**
     * Check that the field is well formed
     * i.e. has a type and an encoding.
     * All macro substition must be done before checkFields is called
     *
     * @exception DTDException IF this field is not well-formed
     * @see MetaTF.Parser.Record#replaceFields
     */
    void checkFields() throws DTDException {
        if(field!=null)
            field.checkFields();
        try {
            int tag = getTag();
            section.addRecordByTag(tag, this);
        } catch(NoTagException e ) {
            //OK
        }
    }

    /**
     * Get the tag of this field
     *
     * @return the tag, if any
     * @exception NoTagException IF this field does not contain a tag
    **/
    public int getTag() throws NoTagException { return field.getTag(); }

    /**
     * Is this field tagged
     *
     * @return whether the field is tagged
    **/
    public boolean isTagged() { return field.isTagged(); }
    
    /** @return the name of this record **/
    public String getName() { return name; }

    /** @return this record's field **/
    public Field getField() { return field; }

/*
    public int numFields() { 
    	if(field==null)
    	    return 0;
    	else if(field instanceof TupleField)
    	    return field.numFields();
    	else return 1;
    }
*/

    /** @return the number of named fields in this record **/
    public int getNumNamedFields() {
        return field==null? 0 : field.getNumNamedFields();
    }

    public void addField(Field f) {
    	if(field==null)
    	    field = f;
    	else if(field instanceof TupleField) 
    	    field.addField(f);
    	else {
    	    TupleField t = new TupleField(name);
    	    t.addField(field);
    	    t.addField(f);
    	    field = t;
    	}
    }

    public void addProperty(String t, String f, String p, Value v, int l) throws DTDException {
    	if(field==null)
    	    error("record "+name+" does not have field "+f);
    	else if(!field.addProperty(t, f, p, v, l))
    	    error("cannot add property "+p+" to field "+f+" of record "+name);
    }

    /** @return the section that this record is defined in **/
    public Section getSection() { return section; }
    
    /**
     * Set the section that this record is defined in
     *
     * @param the name of the section
    **/
    void setSection(Section s) { section = s; }

    /**
     * Compare the name of this record
     *
     * @param the record to be compared
    **/
    public int compareTo(Object r) {
        return name.compareTo(((Record)r).name);
    }

    /**
     * Report an error
     *
     * @param s an error message
     * @exception DTDException ALWAYS
    **/
    private void error(String s) throws DTDException {
        throw new DTDException("l. "+beginLine+": "+s);
    }
    
    /**
     * Dump a representation of this object
     *
     * @param prefix an indentiyfing prefix
     * @return a prettified representation of this object
    **/public String dump(String prefix) {
    	String result = prefix+"event = "+name+":\n";
    	if(field!=null)
    	    result = result + field.dump(prefix+" ");
    	return result+"\n";
    }

    /** Accept the visitor **/
    public Object accept(DTDVisitor visitor, Object data) {
        return visitor.visit(this, data);
    }
}
