/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */
 
package MetaTF.Parser;

import java.util.*;
import MetaTF.Panic;

/**
 * Section
 *
 * @see <a href="metatf.jjt">The MetaTF Grammar</a>
 * @see MetaTF.Parser.Record
 * @see MetaTF.Parser.Field
 * @version rel-1-2 ($Revision: 1.3 $, $Date: 2000/12/15 16:29:55 $),
 * @author Richard Jones
**/


public class Section implements HasFields, HasProperties, Comparable {
  private static final String cvsheader = "$Header: /home/cur/rej/cvs/MetaTF/src/MetaTF/Parser/Section.java,v 1.3 2000/12/15 16:29:55 rej Exp $";

    /** Is this a cached section (default: no) **/
    private boolean iscached = false;

    /** the name of the section, given in the DTD **/
    private final String name;

    /**
     * The number of the section, given in the DTD
     * section O is Global
    **/
    private final int no;

    /** The records defined in this section: Map <record name, Record> **/
    private Map records;

    /** The records defined in this section: Map <tag value, Record> **/
    private Map recordsByTag;
    
    /** The fields defined at the section level **/
    private Map fields;      

    /** The line at which this section's definition in the DTD started **/
    private int beginLine;   
  
    /**
     * Allocate a new object to represent asection
     *
     * @param s the name of the section
     * @param i the number of the section
     * @param b is the section cached?
     * @param the line in the DTD where this section's definition started
     * @exception DTDException IF this section number or name is already used
     */
    Section(String s, int i, boolean b, int l) throws DTDException {
        iscached = b;
        name = s;
        no = i;
        beginLine = l;
        DTD dtd = DTD.instance();
        if(dtd.hasSection(no))
            error("l. "+beginLine+": section "+no+"is already defined");
        if(dtd.hasSection(name))
            error("l. "+beginLine+": section "+name+"is already defined");
        records = new HashMap();
        recordsByTag = new HashMap();
        fields = (no == 0) ?
                     new HashMap() :
                     new HashMap(dtd.getSection(0).getFields());
    }

    /**
     * Allocate a new object to represent asection
     *
     * @param s the name of the section
     * @param i the number of the section
     * @param b is the section cached?
     * @exception DTDException IF this section number is already used
     */
    Section(String s, int i, boolean b) throws DTDException {
        this(s, i, b, 0);
    }
        
    /**
     * Allocate a new object to represent asection
     *
     * @param s the name of the section
     * @param the line in the DTD where this section's definition started
     * @exception DTDException IF this section number is already used
     */
    Section(String s, int i) throws DTDException {
        this(s, i, false, 0);
    }

    /**
     * Add a Record to this section
     * <PRE>
     * 1. if a record of this name is known, remove it
     * 2. do any macro replacement of field names
     * 3. do semantic check on fields
     * 4. finally, add record to section
     * </PRE>
     *
     * @param r the record to add
     * @exception DTDException if this is improper
     */
    void addRecord(Record r) throws DTDException{

        if(!DTD.instance().hasSection(no))
            Panic.panic("The DTD doesn't have section "+no);

        String rname = r.getName();
        if(fields.containsKey(rname))
            fields.remove(rname);       // record name overrides the field name

        r.setSection(this);
        records.put(r.getName(),r);     // this record overrides any previous record name
    }


    // public only because it is an interface method
    public void addProperty(String type, String fname, String p, Value v, int l) throws DTDException { 
    	addField(new NamedField(type, fname, p, v, l));
    }

    /**
     * Get the valeue of the named property if the field is defined at the section level
     *
     * @param fname the name of the field
     * @param p the property
     * @exception DTDException IF fname is not defined at the section level ||
     *                            fname does not have property p
    **/
    public Value getProperty(String fname, String p) throws DTDException {
        if(!fields.containsKey(fname))
            error("section "+name+" does not define field "+fname);
        NamedField f = (NamedField)fields.get(fname);
        if(!f.hasProperty(p))
            error("section "+name+", field "+fname+" does not have property "+p);
        return f.getProperty(p);
    }

    /**
     * Add a Field to this section
     * Simply override old definitions (i.e. don't worry about redefinition)
     *
     * @param f the field to add
     * @exception DTDException IF
     *   f is neither a named not a tag field ||
     *   the DTD doesn't contain this section(!!) 
    **/
    public void addField(Field f) throws DTDException {
        if(!((f instanceof NamedField) || (f instanceof TagField)))
            Panic.panic("l. "+beginLine+", section "+name+
                        ": Section.addField got a field that was neither NamedField not TagField ("+f+")");
        if(!DTD.instance().hasSection(no))
            Panic.panic("l. "+beginLine+": the DTD doesn't have section "+no);
        if(f instanceof NamedField)
            addNamedField((NamedField)f);
        else
            addTagField((TagField)f);
    }

    /**
     * Add a named field to this section
     *
     * @param sfld the named field
     * @exception DTDException
    **/
    private void addNamedField(NamedField sfld) throws DTDException {
        String fname = sfld.getName();
        if(records.containsKey(fname))
            records.remove(fname);
            
        if(!fields.containsKey(fname))
            fields.put(fname,sfld);
        else {
            NamedField oldfield = (NamedField) fields.get(fname);
            Map oldprops = oldfield.getProperties();
            oldprops.putAll(sfld.getProperties());
        }
    }
    
    /**
     * Add a tag field to this section
     *
     * @param sfld the named field
     * @exception DTDException
    **/
    private void addTagField(TagField sfld) throws DTDException {
        if(records.containsKey("tag"))
            records.remove("tag");            
        fields.put("tag",sfld);
    }


    /**
     * Add a Record to the list of 'records by tag' in this section
     *
     * @param tag the tag
     * @param r the record to add
     * @exception DTDException IF a record of this tag has already been added
     */
    void addRecordByTag(int tag, Record r) throws DTDException {
        Integer t = new Integer(tag);
        if(recordsByTag.containsKey(t))
            error("record "+r.getName()+" (tag "+tag+"): duplicate tag "+
                  "(record "+((Record)recordsByTag.get(t)).getName()+")");
        recordsByTag.put(t,r);
    }

    /**
     * Get a record defined in this section
     *
     * @param tag the tag of the record
     * @exception DTDException if this section does not contain a record with tag tag
    **/
    public Record getRecord(int tag) throws DTDException {
        Integer t = new Integer(tag);
        if(!recordsByTag.containsKey(t))
            error("section "+name+"does not have a record with tag "+t);
        return (Record)recordsByTag.get(t);
    }

    /**
     * Get a record defined in this section
     *
     * @param str the name of the record
     * @exception DTDException if this section does not contain a record named str
    **/
    public Record getRecord(String str) throws DTDException {
        if(!records.containsKey(name))
            error("section "+name+"does not have a record with name "+str);
        return (Record)records.get(str);
    }

    /** @return the name of the section **/
    public String getName()    { return name; }

    /** @return the section's number **/
    public int    getNumber()  { return no; }

    /** @return the records defined in this section: Map <record name, Record> **/
    public Map    getRecords() { return records; }

    /** @return the fields defined in this section: Map <field name, Field> **/
    public Map    getFields()  { return fields;}

    /**
     * Does this section define a record?
     *
     * @param name the name of the record
     * @return whether the section contains this record
    **/
    public boolean hasRecord(String name) {
        return records.containsKey(name);
    }
    
    /**
     * Does this section define a record?
     *
     * @param t the tag of the record
     * @return whether the section contains this record
    **/
    public boolean hasRecord(int t) {
        return recordsByTag.containsKey(new Integer(t));
    }

    /** @return the largest tag of any record defined in this section **/
    public int getMaxTag() {
        int max = 0;
        Iterator it = recordsByTag.keySet().iterator();
        while(it.hasNext()) {
            int i = ((Integer)it.next()).intValue();
            if(max<i)
                max = i;
        }
        return max;
    }

    /** @return whether this is a cached section **/
    public boolean isCached() { return iscached; }

    /** @return the number of the line in the DTD where this section's definition started **/
    public int getLine() { return beginLine; }

    /**
     * Compare sections by number
     *
     * @param s the section to compare
     * @return 0 IF section numbers match, 1 IF this section has a higher number, -1 OTHERWISE
    **/
    public int compareTo(Object s) {
        int n = ((Section) s).getNumber();
        if(this.no==n)
            return 0;
        else if(this.no>n)
            return 1;
        else
            return -1;
    }

    /**
     * Report an error
     *
     * @param s an error message
     * @exception DTDException ALWAYS
    **/
    private void error(String s) throws DTDException {
        throw new DTDException("l. "+beginLine+": "+s);
    }

    /**
     * Dump a representation of this section
     *
     * @param prefix an identifying prefix
     * @return a prettified representation of this section
    **/
    public String dump(String prefix) {
        String result = prefix;
        if(iscached)
            result += "cached ";
        result += "Section "+name+" "+no+"\n";
        if(fields!=null) {
            Iterator it = fields.keySet().iterator();
            while(it.hasNext()) {
                String s = (String)it.next();
                Field f = (Field)fields.get(s);
                result = result+f.dump(prefix+" ")+"\n";
            }
        }
        if(records!=null) {
            Iterator it = records.keySet().iterator();
            while(it.hasNext()) {
                String s = (String)it.next();
                Record r = (Record)records.get(s);
                if(r!=null)
                    result = result+r.dump(prefix+" ")+"\n";
            }
        }
        return result;
    }

    /** Accept the visitor **/
    public Object accept(DTDVisitor visitor, Object data) {
        return visitor.visit(this, data);
    }
}
