/*
    $File$
    Date        Author          Changes
    Nov 30 2000 Richard Jones   Created
    
    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */
 
package MetaTF.Runtime;

import MetaTF.MetaTFConstants;
import MetaTF.BadRecordException;
import MetaTF.Panic;
import java.util.*;
import java.io.*;

/**
 * The state of the reader/writer
 *
 * @version rel-1-2 ($Revision: 1.2 $, $Date: 2001/03/29 15:53:24 $),
 * @author Richard Jones
**/
public class MetaTFState {
    
    /** Array of TraceSection, indexed by the section number of DTD sections **/
    static private TraceSection[] sections;

    /** The number of the section currently being read or written **/
    static private int currentSection;

    /**
     * Initialise the reader/writer state
     *
     * @param n the number of sections in the DTD
     * @param f the number of distinct field names
    **/
    public static void init(int n, int f) {
        if(n>0 && f>=0) {
            sections = new TraceSection[n];
            Section.initLookup(n,f);
        } else {
            Panic.panic("Bad values "+n+", "+f+" given to MetaTFState.init");
        }
    }

    /**
     * Initialise a section
     *
     * @param s the section number
     * @param the number of records in section s
    **/
    public static void initSection(int s, int r) {
        if(s>=0 && r>=0) 
           sections[s] = new TraceSection(r);
        else
            Panic.panic("MetaTFState.initSection gets bad values: "+s+", "+r);
    }

    /**
     * Initialise a Record
     *
     * @param s the section number
     * @param r the record tag (i.e. its number)
     * @param f the number of fields
    **/
    public static void initRecord(int s, int r, int f) {
        if(s>=0 && r>=0 && f>=0) 
           sections[s].records[r] = new TraceRecord(f);
        else
            Panic.panic("MetaTFState.initSection gets bad values: "+s+", "+r);
    }

    /**
     * Set a TraceField in the sections/records/fields structure
     *
     * @param s the section number
     * @param r the record number
     * @param f the field number
     * @param val the TraceField
    **/
    public static void setField(int s, int r, int f, int i, TraceField val) {
        try {
            sections[s].records[r].fields[f] = val;
            Section.addLookupField(s, i, val);
        } catch (ArrayIndexOutOfBoundsException e) {
            Panic.panic("MetaTFState.setField got bad values: "+s+", "+r+", "+f+". "+e.getMessage());
        } catch(NullPointerException e) {
            Panic.panic("MetaTFState.setField got bad values: "+s+", "+r+", "+f+". "+e.getMessage());
        }
    }

   

    /**
     * Get TraceRecord
     *
     * @param s the section number
     * @param rtag the tag of the record
     * @return the TraceRecord for this record in this section
    **/
    public static TraceRecord getTraceRecord(int s, int rtag) {
        try {
            return sections[s].records[rtag];
        } catch(ArrayIndexOutOfBoundsException e) {
            Panic.panic("MetaTFState.getTraceRecord gets bad values: "+s+", "+rtag);
            return null;
        }
    }
     
    /**
     * Get TraceField
     *
     * @param s the section number
     * @param rtag the tag of the record
     * @param f the field number
     * @return the TraceField for this record in this section
    **/
    public static TraceField getTraceField(int s, int rtag, int f) {
        try {
            return sections[s].records[rtag].fields[f];
        } catch(ArrayIndexOutOfBoundsException e) {
            Panic.panic("MetaTFState.getTraceField gets bad values: "+s+", "+rtag+" "+f);
            return null;
        }
    }
     

    /**
     * Get the appropriate section
     *
     * @param tag the tag that has just been reaf
    **/
    public static int getCurrentSection(int tag) {
        try {
            return getTraceRecord(currentSection, tag)==null ?
                   0 :
                   currentSection;
        } catch (Exception e) {                                                  
            System.err.println(e.getMessage()+": currentSection="+currentSection+", tag="+Long.toHexString(tag));  
            System.exit(1);                                                      
        }
        return 0; // NEVER REACHED
    }
    
    /**
     * Try to read a metadata record and adjust the state of the reader/writer
     * accordingly.
     *
     * @return the tag
    **/
    public static int readSpecial(LowLevelReader llr) throws BadRecordException, IOException {
      int tag;
      do {
          tag = llr.readTag();                                                 
          if(tag==MetaTFConstants.SectionTag) {                                    
              setSection((int)llr.readNum());                                      
              continue;                                                            
          }                                                                        
          else if(tag==MetaTFConstants.MetadataTag) {                                             
              int recordNumber = 0; // Shut the compiler up!!                                     
              String recordStr = llr.readAny();                                                   
              boolean all = recordStr.charAt(0)==MetaTFConstants.All;
              int field = (int)llr.readNum();                                                     
              int property = (int)llr.readNum();                                                  
              if(!all) {                                                                          
                  recordNumber = Integer.parseInt(recordStr,16);                                  
                  if(recordNumber>sections[currentSection].records.length ||                      
                     sections[currentSection].records[recordNumber]==null)                        
                       throw new BadRecordException("Bad metadata record: Section "+currentSection+" does not have a record with tag "+recordStr);  
                  if(field>sections[currentSection].records[recordNumber].fields.length)          
                       throw new BadRecordException("Bad metadata record: Record (tag "+recordNumber+"), section "+currentSection+", does not have a field "+field);  
              }                                                                                   
              long v1, v2;                                                                        
              TraceField ntf;                                                                      
              switch(property) {                                                                  
                  case MetaTFConstants.None:                                                        
                      if(all)                                                                     
                          Section.changeInterpretation(currentSection, field, MetaTFConstants.None);         
                      else {                                                                      
                          ntf = sections[currentSection].records[recordNumber].fields[field];  
                          ntf.setInterpretationType(MetaTFConstants.None);                          
                      }                                                                           
                      break;                                                                      
                  case MetaTFConstants.BaseOffset:                                                  
                      if(all)                                                                     
                          Section.changeInterpretation(currentSection, field, MetaTFConstants.BaseOffset, llr.readNum());   
                      else {                                                                      
                          try {                                                                   
                              ntf = sections[currentSection].records[recordNumber].fields[field];  
                              ntf.setInterpretationType(MetaTFConstants.BaseOffset);                    
                              ntf.setValue(llr.readNum());                                
                          } catch (ClassCastException e) {                                        
                              Panic.panic("Cannot set a baseOffset encoding for a non-numeric type"); 
                          }                                                                       
                      }                                                                           
                      break;                                                                      
                  case MetaTFConstants.Delta:                                                       
                      if(all)                                                                     
                          Section.changeInterpretation(currentSection, field, MetaTFConstants.Delta, llr.readNum());        
                      else {                                                                      
                          try {                                                                   
                              ntf = sections[currentSection].records[recordNumber].fields[field];  
                              ntf.setInterpretationType(MetaTFConstants.Delta);                             
                              ntf.setValue(llr.readNum());                                    
                          } catch (ClassCastException e) {                                        
                              Panic.panic("Cannot set a delta encoding for a non-numeric type"); 
                          }                                                                       
                      }                                                                           
                      break;                                                                      
                  case MetaTFConstants.Default:                                                   
                      if(all)                                                                     
                          Section.changeInterpretation(currentSection, field, MetaTFConstants.Default, llr.readNum());     
                      else {                                                                      
                          ntf = sections[currentSection].records[recordNumber].fields[field];     
                          ntf.setInterpretationType(MetaTFConstants.Default);                     
                          if(ntf instanceof StringTraceField)                                     
                              ntf.setValue(llr.readString());                                     
                          else                                                                    
                              ntf.setValue(llr.readNum());                                        
                      }                                                                           
                      break;                                                                      
                  case MetaTFConstants.SectionOffset:                                             
                      if(all)                                                                     
                          Section.changeInterpretation(currentSection, field, MetaTFConstants.SectionOffset, llr.readNum()); 
                      else {                                                                      
                          try {                                                                   
                              ntf = sections[currentSection].records[recordNumber].fields[field];  
                              ntf.setInterpretationType(MetaTFConstants.SectionOffset);                 
//                            ntf.setValue(llr.readNum());                                
                          } catch (ClassCastException e) {                                        
                              Panic.panic("Can only set sectionOffset encoding for index types"); 
                          }                                                                       
                      }                                                                           
                      break;                                                                      
                   case MetaTFConstants.Stride:                                                      
                       v1 = llr.readNum();                                                         
                       v2 = llr.readNum();                                                         
                       if(all)                                                                     
                           Section.changeInterpretation(currentSection, field, MetaTFConstants.Stride,v1, v2);               
                       else {                                                                      
                            try {                                                                   
                                ntf = sections[currentSection].records[recordNumber].fields[field];  
                                ntf.setInterpretationType(MetaTFConstants.Stride);                        
                                ntf.setValue(v1);                                           
                                ntf.setValue2(v2);                                           
                            } catch (ClassCastException e) {                                        
                                Panic.panic("Cannot set a stride encoding for a non-numeric type"); 
                            }                                                                       
                       }                                                                           
                       break;                                                                      
                    case MetaTFConstants.SectionStride:                                               
                        v1 = llr.readNum();                                                         
                        v2 = llr.readNum();                                                         
                        if(all)                                                                     
                            Section.changeInterpretation(currentSection, field, MetaTFConstants.SectionStride,v1, v2);        
                        else {                                                                      
                            try {                                                                   
                                ntf = sections[currentSection].records[recordNumber].fields[field];  
                                ntf.setInterpretationType(MetaTFConstants.SectionStride);                 
                                ntf.setValue(v1);                                           
                                ntf.setValue2(v2);                                           
                            } catch (ClassCastException e) {                                        
                                Panic.panic("Can only set sectionStride encoding for index types"); 
                            }                                                                       
                        }                                                                           
                        break;                                                                      
              }                                                                               
              continue;                                                                       
          }

          else return tag; // user-defined tag
              
      } while(true);
    }
                                                                 

    /**
     * High-level read primitive to read a number according to the field's
     * encoding held in a NumTraceField
     *
     * @param tf the NumTraceField
     * @exception BadRecordException IF the encoding is MetaTFConstants.SectionOffset
     *                                  or MetaTFConstants.SectionStride
     * @exception IOException IF the read fails
    **/                                                                                    
    public static long readNum(LowLevelReader llr, NumTraceField tf) throws BadRecordException, IOException  {               
        long i;                                                                            
        switch(tf.getInterpretationType()) {                                               
            case MetaTFConstants.None:                                                     
                return llr.readNum();                                                      
            case MetaTFConstants.BaseOffset:                                               
                return tf.getValue() + llr.readNum();                                     
            case MetaTFConstants.Delta:                                                    
                i = tf.getValue()+llr.readNum();                                          
                tf.setValue(i);                                                           
                return i;                                                                  
            case MetaTFConstants.Default:                                                  
                return tf.getValue();                                                     
            case MetaTFConstants.SectionOffset:                                            
                Panic.panic("readNum: cannot read a number into a sectionOffset");             
                break;                                                                     
            case MetaTFConstants.Stride:                                                   
                i = tf.getValue();                                       
                tf.setValue(i + tf.getValue2());                                                           
                return i;                                                                  
            case MetaTFConstants.SectionStride:                                            
                Panic.panic("readNum: cannot read a number into a sectionOffset");             
                break;                                                                     
        }                                                                                  
        return 0; // NOT REACHED                                                           
    }                                                                                      
                                                                                           
    /**
     * High-level read primitive to read a string according to the field's
     * encoding held in a StringTraceField
     *
     * @param tf the StringTraceField
     * @exception BadRecordException IF the encoding is neither MetaTFConstants.None
     *                                  nor MetaTFConstants.Default
     * @exception IOException IF the read fails
    **/                                                                                    
    public static String readString(LowLevelReader llr, StringTraceField tf) throws BadRecordException, IOException {        
        switch(tf.getInterpretationType()) {                                                    
            case MetaTFConstants.None:                                                     
                return llr.readString();                                                   
            case MetaTFConstants.Default:                                                  
                return tf.getValue();                                                      
            case MetaTFConstants.BaseOffset:                                               
            case MetaTFConstants.Delta:                                                    
            case MetaTFConstants.SectionOffset:                                            
            case MetaTFConstants.Stride:                                                   
            case MetaTFConstants.SectionStride:                                            
                Panic.panic("Cannot read numeric value into a string");                        
                break;                                                                     
        }                                                                                  
        return null; // NOT REACHED                                                        
    }

    /**
     * Reads a string with assumed encoding of none
     * e.g. used to read Comments
    **/
    public static String readln(LowLevelReader llr) throws BadRecordException, IOException {
        return llr.readln();
    }
                                                                                           
    /**
     * High-level read primitive to read an indexed value according to the field's
     * encoding held in a IndexTraceField
     *
     * @param tf the IndexTraceField
     * @exception BadRecordException IF the encoding is neither MetaTFConstants.SectionOffset
     *                                  MetaTFConstants.SectionStride nor MetaTFConstants.Default
     * @exception IOException IF the read fails
    **/                                                                                    
    public static Index readIndex(LowLevelReader llr, IndexTraceField tf) throws BadRecordException, IOException {           
        List recs;                                                                         
        switch(tf.getInterpretationType()) {                                                    
            case MetaTFConstants.None:                                                     
            case MetaTFConstants.BaseOffset:                                               
            case MetaTFConstants.Delta:                                                    
            case MetaTFConstants.Stride:                                                   
                Panic.panic("Cannot read numeric value into an index");                        
                break;                                                                     
            case MetaTFConstants.Default:                                                  
                return null;                                                               
            case MetaTFConstants.SectionOffset:                                            
                int o = (int)llr.readNum();                                                
                recs = tf.getRecords();                                                    
                if(o<0 || o>=recs.size())                                                  
                    throw new BadRecordException("Invalid sectionOffset "+o);            
                return new Index((Record)recs.get(o), o);                                  
            case MetaTFConstants.SectionStride:                                            
                int c = (int)tf.getValue();                                                
                int i = c + tf.getValue2();                                                
                tf.setValue(i);                                                            
                recs = tf.getRecords();                                                    
                if(c<0 || c>=recs.size())                                                  
                    throw new BadRecordException("Invalid sectionStride "+i);            
                return new Index((Record)recs.get(c), c);                                  
        }                                                                                  
        return null; // NOT REACHED                                                        
    }                                                                                      

    /**                                                                            
     * Change the current section in the reader/writer                             
     *                                                                             
     * @param i the new section which must be valid                                
     * @exception BadRecordException IF i is an invalid section number             
    **/                                                                            
    public static void setSection(int i) throws BadRecordException {                      
        try {                                                                      
            if(sections[i]!=null)                                                  
                currentSection = i;                                                
            else                                                                   
                throw new BadRecordException("Bad section metadata "+i);         
        } catch(ArrayIndexOutOfBoundsException e) {                                
            throw new BadRecordException("Bad section metadata "+i);             
        }                                                                          
    }

    
    /**                                                                            
     * Write section metadata to change current section                            
     * after changing the current section in the reader/writer                     
     *                                                                             
     * @param i the new section which must be valid                                
     * @exception BadRecordException IF i is an invalid section number             
     * @exception IOException IF the write fails                                   
    **/                                                                            
    public static void putSection(LowLevelWriter llw, int i) throws BadRecordException, IOException {               
        setSection(i);                                                             
        llw.write(MetaTFConstants.Section);                                       
        llw.write(i);
        llw.writeln();
    }
}
